@@ -1,28 +1,23 @@
 import logging
 import shutil
-from copy import deepcopy
-from typing import Dict, Any, Type
-
 import h5py
+import numpy
+
+from copy import deepcopy
+from typing import Dict, Any
+
 from Bio import SeqIO
 from pandas import read_csv, DataFrame
 from tqdm import tqdm
 
-from bio_embeddings.embed import (
-    ProtTransAlbertBFDEmbedder,
-    ProtTransBertBFDEmbedder,
-    EmbedderInterface,
-    SeqVecEmbedder,
-    ProtTransXLNetUniRef100Embedder,
-    UniRepEmbedder,
-)
+from bio_embeddings.embed import name_to_embedder, EmbedderInterface
 from bio_embeddings.utilities import (
+    FileManagerInterface,
     InvalidParameterError,
-    get_model_file,
     check_required,
     get_file_manager,
     get_model_directories_from_zip,
-    FileManagerInterface,
+    get_model_file,
 )
 from bio_embeddings.utilities.backports import nullcontext
 
@@ -58,6 +53,7 @@ def _print_expected_file_sizes(
 
         required_space_in_MB += reduced_embeddings_file_size_in_MB
 
+    # TODO: calculate size of transformed embeddings?
     if not (result_kwargs.get("reduce") is True and result_kwargs.get("discard_per_amino_acid_embeddings") is True):
         logger.info(f"The minimum expected size for the embedding_file is {embeddings_file_size_in_MB:.3f}MB.")
 
@@ -117,9 +113,10 @@ def _get_embeddings_file_context(
     result_kwargs.setdefault("discard_per_amino_acid_embeddings", False)
 
     if result_kwargs["discard_per_amino_acid_embeddings"] is True:
-        if result_kwargs["reduce"] is False:
+        if result_kwargs.get("reduce", False) is False and result_kwargs.get("embeddings_transformer_function") is None:
             raise InvalidParameterError(
-                "Cannot have discard_per_amino_acid_embeddings=True and reduce=False. Both must be True."
+                "Cannot only have discard_per_amino_acid_embeddings: True. "
+                "Either also set `reduce: True` or define an `embeddings_transformer_function`, or both."
             )
         return nullcontext()
     else:
@@ -133,10 +130,73 @@ def _get_embeddings_file_context(
         return h5py.File(embeddings_file_path, "w")
 
 
+def _get_transformed_embeddings_file_context(
+    file_manager: FileManagerInterface, result_kwargs: Dict[str, Any]
+):
+    """
+    :param file_manager: The FileManager derived class which will be used to create the file
+    :param result_kwargs: A dictionary which will be updated in-place to include the path to the newly created file
+
+    :return: a file context
+    """
+
+    result_kwargs.setdefault("embeddings_transformer_function", None)
+
+    if result_kwargs["embeddings_transformer_function"] is not None:
+        transformed_embeddings_file_path = file_manager.create_file(
+            result_kwargs.get("prefix"),
+            result_kwargs.get("stage_name"),
+            "transformed_embeddings_file",
+            extension=".h5",
+        )
+        result_kwargs["transformed_embeddings_file"] = transformed_embeddings_file_path
+        return h5py.File(transformed_embeddings_file_path, "w")
+
+    return nullcontext()
+
+
+def _check_transform_embeddings_function(embedder: EmbedderInterface, result_kwargs: Dict[str, Any]):
+
+    result_kwargs.setdefault("embeddings_transformer_function", None)
+
+    if result_kwargs["embeddings_transformer_function"] is not None:
+        try:
+            transform_function = eval(result_kwargs["embeddings_transformer_function"], {}, {"np": numpy})
+        except TypeError:
+            raise InvalidParameterError(f"`embeddings_transformer_function` must be callable! \n"
+                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
+                                        f"Most likely you want a lambda function.")
+
+        if not callable(transform_function):
+            raise InvalidParameterError(f"`embeddings_transformer_function` must be callable! \n"
+                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
+                                        f"Most likely you want a lambda function.")
+
+        template_embedding = embedder.embed("SEQVENCE")
+
+        # Check that it works in principle
+        try:
+            transformed_template_embedding = transform_function(template_embedding)
+        except:
+            raise InvalidParameterError(f"`embeddings_transformer_function` must be valid callable! \n"
+                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
+                                        f"This function excepts when processing an embedding.")
+
+        # Check that return can be cast to np.array
+        try:
+            numpy.array(transformed_template_embedding)
+        except:
+            raise InvalidParameterError(f"`embeddings_transformer_function` must be valid callable "
+                                        f"returning numpy array compatible object! \n"
+                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
+                                        f"This function excepts when processing an embedding.")
+
+
 def embed_and_write_batched(
     embedder: EmbedderInterface,
     file_manager: FileManagerInterface,
     result_kwargs: Dict[str, Any],
+    half_precision: bool = False
 ) -> Dict[str, Any]:
     """ The shared code between the SeqVec, Albert, Bert and XLNet pipelines """
     # Lazy fasta file reader. The mapping file contains the corresponding ids in the same order
@@ -152,93 +212,77 @@ def embed_and_write_batched(
     # Print the minimum required file sizes
     _print_expected_file_sizes(embedder, mapping_file, result_kwargs)
 
+    # Get transformer function, if available
+    transform_function = result_kwargs.get("embeddings_transformer_function", None)
+
+    if transform_function:
+        transform_function = eval(transform_function, {}, {"np": numpy})
+
     # Open embedding files or null contexts and iteratively save embeddings to file
     with _get_embeddings_file_context(
         file_manager, result_kwargs
     ) as embeddings_file, _get_reduced_embeddings_file_context(
         file_manager, result_kwargs
-    ) as reduced_embeddings_file:
+    ) as reduced_embeddings_file, _get_transformed_embeddings_file_context(
+        file_manager, result_kwargs
+    ) as transformed_embeddings_file:
         embedding_generator = embedder.embed_many(
             sequences, result_kwargs.get("max_amino_acids")
         )
-        for sequence_id, embedding in zip(
-            mapping_file.index, tqdm(embedding_generator, total=len(mapping_file))
+        for sequence_id, original_id, embedding in zip(
+            mapping_file.index,
+            mapping_file["original_id"],
+            tqdm(embedding_generator, total=len(mapping_file))
         ):
+            # embedding: numpy.ndarray
+            if half_precision:
+                embedding = embedding.astype(numpy.float16)
             if result_kwargs.get("discard_per_amino_acid_embeddings") is False:
-                embeddings_file.create_dataset(sequence_id, data=embedding)
-
+                dataset = embeddings_file.create_dataset(sequence_id, data=embedding)
+                dataset.attrs["original_id"] = original_id
             if result_kwargs.get("reduce") is True:
-                reduced_embeddings_file.create_dataset(
+                dataset = reduced_embeddings_file.create_dataset(
                     sequence_id, data=embedder.reduce_per_protein(embedding)
                 )
+                dataset.attrs["original_id"] = original_id
+            if transform_function:
+                dataset = transformed_embeddings_file.create_dataset(
+                    sequence_id, data=numpy.array(transform_function(embedding))
+                )
+                dataset.attrs["original_id"] = original_id
+
     return result_kwargs
 
 
-def seqvec(**kwargs) -> Dict[str, Any]:
-    necessary_files = ["weights_file", "options_file"]
-    result_kwargs = deepcopy(kwargs)
-    file_manager = get_file_manager(**kwargs)
+# Some of this might not be available when installed without the `all` extra
+ALL_PROTOCOLS = [
+    "bepler",
+    "cpcprot",
+    "esm",
+    "esm1b",
+    "plus_rnn",
+    "prottrans_albert_bfd",
+    "prottrans_bert_bfd",
+    "prottrans_t5_bfd",
+    "prottrans_xlnet_uniref100",
+    "seqvec",
+    "unirep",
+]
 
-    # Initialize pipeline and model specific options:
-    result_kwargs.setdefault("max_amino_acids", 15000)
-
-    # Download necessary files if needed
-    for file in necessary_files:
-        if not result_kwargs.get(file):
-            result_kwargs[file] = get_model_file(model="seqvec", file=file)
-
-    embedder = SeqVecEmbedder(**result_kwargs)
-    return embed_and_write_batched(embedder, file_manager, result_kwargs)
-
-
-def transformer(
-    embedder_class: Type[EmbedderInterface], max_amino_acids_default: int, **kwargs
-):
-    result_kwargs = deepcopy(kwargs)
-    file_manager = get_file_manager(**kwargs)
-    result_kwargs.setdefault("max_amino_acids", max_amino_acids_default)
-
-    necessary_directories = ["model_directory"]
-    for directory in necessary_directories:
-        if not result_kwargs.get(directory):
-            result_kwargs[directory] = get_model_directories_from_zip(
-                model=embedder_class.name, directory=directory
-            )
-
-    embedder = embedder_class(**result_kwargs)
-    return embed_and_write_batched(embedder, file_manager, result_kwargs)
-
-
-def prottrans_albert(**kwargs):
-    return transformer(ProtTransAlbertBFDEmbedder, 3035, **kwargs)
-
-
-def prottrans_bert_bfd(**kwargs):
-    return transformer(ProtTransBertBFDEmbedder, 6024, **kwargs)
-
-
-def prottrans_xlnet(**kwargs):
-    return transformer(ProtTransXLNetUniRef100Embedder, 4000, **kwargs)
-
-
-def unirep(**kwargs) -> Dict[str, Any]:
-    if kwargs.get("use_cpu") is not None:
-        raise InvalidParameterError("UniRep does not support configuring `use_cpu`")
-    result_kwargs = deepcopy(kwargs)
-    file_manager = get_file_manager(**kwargs)
-    embedder = UniRepEmbedder(**result_kwargs)
-    # We don't actually batch with UniRep, but embed_and_write_batched
-    # works anyway since UniRepEmbedder still implements `embed_many`
-    return embed_and_write_batched(embedder, file_manager, result_kwargs)
-
-
-# list of available embedding protocols
-PROTOCOLS = {
-    "seqvec": seqvec,
-    "prottrans_albert_bfd": prottrans_albert,
-    "prottrans_bert_bfd": prottrans_bert_bfd,
-    "prottrans_xlnet_uniref100": prottrans_xlnet,
-    "unirep": unirep,
+# TODO: 10000 is a random guess
+# There remainder was measured for a GTX 1080 with 8GB memory
+DEFAULT_MAX_AMINO_ACIDS = {
+    "bepler": 10000,
+    "cpcprot": 10000,
+    "esm": 10000,
+    "esm1b": 10000,
+    "plus_rnn": 10000,
+    "prottrans_albert_bfd": 3035,
+    "prottrans_bert_bfd": 6024,
+    "prottrans_t5_bfd": 1000,
+    "prottrans_xlnet_uniref100": 4000,
+    "seqvec": 15000,
+    "unirep": 10000,
 }
 
 
@@ -264,11 +308,42 @@ def run(**kwargs):
         ["protocol", "prefix", "stage_name", "remapped_sequences_file", "mapping_file"],
     )
 
-    if kwargs["protocol"] not in PROTOCOLS:
+    if kwargs["protocol"] not in name_to_embedder:
+        if kwargs["protocol"] in ALL_PROTOCOLS:
+            raise InvalidParameterError(
+                f"The extra for the protocol {kwargs['protocol']} is missing. "
+                "See https://docs.bioembeddings.com/#installation on how to install all extras"
+            )
         raise InvalidParameterError(
             "Invalid protocol selection: {}. Valid protocols are: {}".format(
-                kwargs["protocol"], ", ".join(PROTOCOLS.keys())
+                kwargs["protocol"], ", ".join(name_to_embedder.keys())
             )
         )
 
-    return PROTOCOLS[kwargs["protocol"]](**kwargs)
+    embedder_class = name_to_embedder[kwargs["protocol"]]
+
+    if kwargs["protocol"] == "unirep" and kwargs.get("use_cpu") is not None:
+        raise InvalidParameterError("UniRep does not support configuring `use_cpu`")
+
+    result_kwargs = deepcopy(kwargs)
+
+    # Download necessary files if needed
+    # noinspection PyProtectedMember
+    for file in embedder_class._necessary_files:
+        if not result_kwargs.get(file):
+            result_kwargs[file] = get_model_file(model=embedder_class.name, file=file)
+
+    # noinspection PyProtectedMember
+    for directory in embedder_class._necessary_directories:
+        if not result_kwargs.get(directory):
+            result_kwargs[directory] = get_model_directories_from_zip(
+                model=embedder_class.name, directory=directory
+            )
+
+    result_kwargs.setdefault("max_amino_acids", DEFAULT_MAX_AMINO_ACIDS[kwargs["protocol"]])
+
+    file_manager = get_file_manager(**kwargs)
+    embedder: EmbedderInterface = embedder_class(**result_kwargs)
+    _check_transform_embeddings_function(embedder, result_kwargs)
+
+    return embed_and_write_batched(embedder, file_manager, result_kwargs, kwargs.get("half_precision", False))
